import { Channel } from './channel';
import { ChatKittyError } from './error';
import { ChatKittyFile, ChatKittyUploadProgressListener, CreateChatKittyFileProperties } from './file';
import { ChatKittyPaginator } from './pagination';
import { ReactionSummary } from './reaction';
import { ChatKittyFailedResult, ChatKittySucceededResult } from './result';
import { Thread } from './thread';
import { User } from './user';
export type Message = SystemMessage | UserMessage;
export type SystemMessage = TextSystemMessage | FileSystemMessage;
export type UserMessage = TextUserMessage | FileUserMessage;
export type TextMessage = TextSystemMessage | TextUserMessage;
export type FileMessage = FileSystemMessage | FileUserMessage;
export interface BaseMessage {
    id: number;
    type: string;
    channelId: number;
    createdTime: string;
    groupTag?: string;
    reactions?: ReactionSummary[];
    repliesCount?: number;
    properties: unknown;
    /** @internal */
    _relays: MessageRelays;
    /** @internal */
    _actions: MessageActions;
    /** @internal */
    _streams: MessageStreams;
}
export type BaseTextMessage = BaseMessage & {
    body: string;
    links: MessageLink[];
    mentions?: MessageMention[];
};
export type BaseFileMessage = BaseMessage & {
    file: ChatKittyFile;
};
export type BaseUserMessage = BaseMessage & {
    user: User;
};
export type TextSystemMessage = BaseTextMessage & {
    type: 'SYSTEM_TEXT';
};
export type FileSystemMessage = BaseFileMessage & {
    type: 'SYSTEM_FILE';
};
export type TextUserMessage = BaseTextMessage & BaseUserMessage & {
    type: 'TEXT';
};
export type FileUserMessage = BaseFileMessage & BaseUserMessage & {
    type: 'FILE';
};
export declare class MessageLink {
    source: string;
    startPosition: number;
    endPosition: number;
    preview?: MessageLinkPreview;
}
export declare class MessageLinkPreview {
    url: string;
    title: string;
    image: MessageLinkPreviewImage;
    description?: string;
    siteName?: string;
}
export declare class MessageLinkPreviewImage {
    source: string;
}
export type MessageMention = ChannelMessageMention | UserMessageMention;
export interface BaseMessageMention {
    type: string;
    tag: string;
    startPosition: number;
    endPosition: number;
}
export type ChannelMessageMention = BaseMessageMention & {
    channel: Channel;
};
export type UserMessageMention = BaseMessageMention & {
    user: User;
};
declare class MessageRelays {
    self: string;
    channel: string;
    parent?: string;
    readReceipts: string;
    repliesCount: string;
    replies: string;
    reactions: string;
}
declare class MessageActions {
    read: string;
    unread: string;
    reply: string;
    deleteForMe: string;
    delete: string;
    react: string;
    edit: string;
    removeReaction: string;
    updateProperties: string;
}
declare class MessageStreams {
    replies: string;
}
export declare function isTextMessage(message: Message): message is TextMessage;
export declare function isFileMessage(message: Message): message is FileMessage;
export declare function isUserMessage(message: Message): message is UserMessage;
export declare function isSystemMessage(message: Message): message is SystemMessage;
export type ListMessagesRequest = ListUsersMessagesRequest | ListChannelMessagesRequest | ListMessageRepliesRequest;
export declare class ListUsersMessagesRequest {
    filter?: ListUsersMessagesFilter;
}
export declare class ListChannelMessagesRequest {
    channel: Channel;
    filter?: ListChannelMessagesFilter;
}
export declare class ListMessageRepliesRequest {
    message: Message;
}
export declare class ListUsersMessagesFilter {
    type: 'FILE' | undefined;
}
export declare class ListChannelMessagesFilter {
    mainThread: boolean;
    type: 'FILE' | undefined;
}
export declare class RetrieveLastReadMessageRequest {
    channel: Channel;
    username: string;
}
export type ListMessagesResult = ListMessagesSucceededResult | ChatKittyFailedResult;
export declare class ListMessagesSucceededResult extends ChatKittySucceededResult {
    paginator: ChatKittyPaginator<Message>;
    constructor(paginator: ChatKittyPaginator<Message>);
}
export type RetrieveLastReadMessageResult = RetrieveLastReadMessageSucceededResult | ChatKittyFailedResult;
export declare class RetrieveLastReadMessageSucceededResult extends ChatKittySucceededResult {
    message?: Message | undefined;
    constructor(message?: Message | undefined);
}
export declare class ReadMessageRequest {
    message: Message;
}
export declare class UnreadMessageRequest {
    message: Message;
}
export type ReadMessageResult = ReadMessageSucceededResult | ChatKittyFailedResult;
export declare class ReadMessageSucceededResult extends ChatKittySucceededResult {
    message: Message;
    constructor(message: Message);
}
export type UnreadMessageResult = UnreadMessageSucceededResult | ChatKittyFailedResult;
export declare class UnreadMessageSucceededResult extends ChatKittySucceededResult {
    message: Message;
    constructor(message: Message);
}
export declare class EditMessageRequest {
    message: Message;
    body: string;
}
export type EditMessageResult = EditedMessageSucceededResult | ChatKittyFailedResult;
export declare class EditedMessageSucceededResult extends ChatKittySucceededResult {
    message: Message;
    constructor(message: Message);
}
export declare class UpdateMessagePropertiesRequest {
    message: Message;
    properties: unknown;
}
export type UpdateMessagePropertiesResult = UpdatedMessagePropertiesSucceededResult | ChatKittyFailedResult;
export declare class UpdatedMessagePropertiesSucceededResult extends ChatKittySucceededResult {
    message: Message;
    constructor(message: Message);
}
export declare class DeleteMessageRequest {
    message: Message;
}
export type DeleteMessageResult = DeleteMessageSucceededResult | ChatKittyFailedResult;
export declare class DeleteMessageSucceededResult extends ChatKittySucceededResult {
    message: Message;
    constructor(message: Message);
}
export type SendMessageRequest = SendTextMessageRequest | SendFileMessageRequest;
export type SendChannelMessageRequest = {
    channel: Channel;
};
export type SendMessageReplyRequest = {
    message: Message;
};
export type SendThreadMessageRequest = {
    thread: Thread;
};
export type SendTextMessageRequest = (SendChannelMessageRequest | SendMessageReplyRequest | SendThreadMessageRequest) & {
    body: string;
    groupTag?: string;
    properties?: unknown;
};
export type SendFileMessageRequest = (SendChannelMessageRequest | SendMessageReplyRequest | SendThreadMessageRequest) & {
    file: CreateChatKittyFileProperties;
    groupTag?: string;
    properties?: unknown;
    progressListener?: ChatKittyUploadProgressListener;
};
export type SendMessageResult = SentMessageResult | ChatKittyFailedResult;
export type SentMessageResult = SentTextMessageResult | SentFileMessageResult;
export declare class SentTextMessageResult extends ChatKittySucceededResult {
    message: TextUserMessage;
    constructor(message: TextUserMessage);
}
export declare class SentFileMessageResult extends ChatKittySucceededResult {
    message: FileUserMessage;
    constructor(message: FileUserMessage);
}
export declare function sentTextMessage(result: SentMessageResult): result is SentTextMessageResult;
export declare function sentFileMessage(result: SentMessageResult): result is SentFileMessageResult;
export declare class CountUnreadMessagesRequest {
    channel: Channel;
}
export declare class CountMessageRepliesRequest {
    message: Message;
}
export declare class RetrieveMessageChannelRequest {
    message: Message;
}
export type RetrieveMessageChannelResult = RetrieveMessageChannelSucceededResult | ChatKittyFailedResult;
export declare class RetrieveMessageChannelSucceededResult extends ChatKittySucceededResult {
    channel: Channel;
    constructor(channel: Channel);
}
export declare class RetrieveMessageParentRequest {
    message: Message;
}
export type RetrieveMessageParentResult = RetrieveMessageParentSucceededResult | ChatKittyFailedResult;
export declare class RetrieveMessageParentSucceededResult extends ChatKittySucceededResult {
    message: Message;
    constructor(message: Message);
}
export declare class MessageNotAReplyError extends ChatKittyError {
    messageModel: Message;
    constructor(messageModel: Message);
}
export {};
//# sourceMappingURL=message.d.ts.map